from .json_schema.parse import parse as parse_json_schema
from .regex.parse import parse as parse_regex
from .xml_schema.parse import parse as parse_xml_schema
from .grammar.convert import convert as parse_grammar
